#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : hubdiags.py
# Contents: create diagrams for hubness phenomenon
# Author  : Christian Borgelt
# History : 2012.01.23 file created
#-----------------------------------------------------------------------
from sys  import argv, stdout
from math import floor

#-----------------------------------------------------------------------

colors = [ 'red', '0.7green', 'blue', 'black',
           'red', '0.7green', 'blue',
           'blue', 'red' ]
styles = [ '', '', '', '', '', '', '', '', '' ]

#-----------------------------------------------------------------------

def identity (x):
    return x

#-----------------------------------------------------------------------

def minus1 (x):
    return x-1

#-----------------------------------------------------------------------

def cubes (x):
    if x < 1: return 1
    return int(floor(x))

#-----------------------------------------------------------------------

def csize (x):
    if x < 1: return 0
    return 100 -int(100 *(x -floor(x)) +0.5)

#-----------------------------------------------------------------------

def header (out, pro):
    out.write('verbatimtex\n')
    out.write('\\documentclass{svmult}\n')
    out.write('\\begin{document}\n')
    out.write('\\footnotesize\n')
    out.write('etex\n\n')
    out.write('prologues := ' +str(pro) +';\n\n')
    out.write('input linechart;\n\n')

#-----------------------------------------------------------------------

def title (out, text):
    out.write('  linechart.draw_title(0, btex ' +text +' etex);\n')

#-----------------------------------------------------------------------

def data (out, s='c', n=1000, d=2, k=5, g=0, target=0, xmap=identity):
    params = dict([]); loop = ()
    if isinstance(s, [].__class__): loop = ('s', s)
    else:                           params['s'] = s
    if isinstance(n, [].__class__): loop = ('n', n)
    else:                           params['n'] = n
    if isinstance(d, [].__class__): loop = ('d', d)
    else:                           params['d'] = d
    if isinstance(k, [].__class__): loop = ('k', k)
    else:                           params['k'] = k
    if isinstance(g, [].__class__): loop = ('g', g)
    else:                           params['g'] = g
    out.write('  linechart.pfill(data');
    for x in loop[1]:
        params[loop[0]] = x
        fname =  's' +str(params['s']) \
              + '_n' +str(params['n']) \
              + '_d' +str(params['d']) \
              + '_k' +str(params['k']) \
              + '_g' +str(params['g']) +'.txt'
        measures = ()
        with open(fname, 'rt') as inp:
            for line in inp:
                if measures == '':
                    measures = line.split(); break
                elif line[0:4] == '----':
                    measures = ''
            t = float(measures[target])
            if t < 0: t = 0
            out.write(', ({0:.5f}'.format(xmap(x) +0.0))
            out.write(',{0:.5f}'.format(t) +')')
    out.write(');\n')

#-----------------------------------------------------------------------

def curve (out, text, pos, color='red', style=''):
    out.write('  linechart.draw_curve(data, '
             +'btex \\scriptsize ' +text +' etex, '
             +str(pos) +', 0, ' +color +', ' +style +');\n')

#-----------------------------------------------------------------------

def legend (out, pos, text):
    out.write('  linechart.draw_legend(' +str(pos) +'*mm, ')
    if text != '': out.write('btex \\scriptsize ' +text +' etex);\n')
    else:          out.write('"");\n')

#-----------------------------------------------------------------------

def footer (out):
    out.write('end.\n')

#-----------------------------------------------------------------------

def hubs_diags(out, fid):
    out.write('vardef init_hubs (expr ht, txt) =\n')
    out.write('  pair data[];\n')
    out.write('  linechart.set_defaults("ptmr8r");\n')
    out.write('  linechart.legend.linelen  := 3mm;\n')
    out.write('  linechart.legend.itemskip := '
              + '0.9linechart.legend.itemskip;\n')
    out.write('  linechart.set_units(44mm/104, 26mm/ht);\n')
    out.write('  linechart.set_ranges((0,104), (0,ht));\n')
    out.write('  linechart.draw_xticks(0, 20, 40, 60, 80, 100);\n')
    out.write('  linechart.draw_xlabel(0, '
                   +'btex number of dimensions $n$ etex);\n')
    out.write('  if     ht >= 14:\n');
    out.write('    linechart.y.alabel.offset := (0,5.5mm);\n')
    out.write('    linechart.draw_yticks(0, 5, 10, 15, 20);\n')
    out.write('  elseif ht >= 12:\n');
    out.write('    linechart.y.alabel.offset := (0,4.0mm);\n')
    out.write('    linechart.draw_yticks(0, 3, 6, 9, 12);\n')
    out.write('  else:\n')
    out.write('    linechart.y.alabel.offset := (0,3.2mm);\n')
    out.write('    linechart.draw_yticks(0, 2, 4, 6, 8, 10);\n')
    out.write('  fi\n')
    out.write('  linechart.draw_frame;\n')
    out.write('  linechart.draw_title(2, txt);\n')
    out.write('enddef;\n\n')

    dimlist = [ 2, 3, 5, 10, 15, 20, 30, 40, 50, 60, 80, 100 ]
    spdlist = [ x+1 for x in dimlist ]

    out.write('beginfig(' +str(fid) +');\n')
    out.write('  init_hubs(11.4, '
             +'btex 10 nearest neighbors, 10000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex skewness $\gamma$ etex);\n')
    data (out, 's', 10000, spdlist, 10, 0, 0, minus1)
    curve(out, 'sphere', 3, colors[3], styles[3])
    data (out, 'b', 10000, dimlist, 10, 0, 0, identity)
    curve(out, 'ball',   2, colors[2], styles[2])
    data (out, 'c', 10000, dimlist, 10, 0, 0, identity)
    curve(out, 'cube',   1, colors[1], styles[1])
    data (out, 'n', 10000, dimlist, 10, 0, 0, identity)
    curve(out, 'normal', 0, colors[0], styles[0])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+1) +');\n')
    out.write('  init_hubs(11.4, '
             +'btex 10 nearest neighbors, 20000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex skewness $\gamma$ etex);\n')
    data (out, 's', 20000, spdlist, 10, 0, 0, minus1)
    curve(out, 'sphere', 3, colors[3], styles[3])
    data (out, 'b', 20000, dimlist, 10, 0, 0, identity)
    curve(out, 'ball',   2, colors[2], styles[2])
    data (out, 'c', 20000, dimlist, 10, 0, 0, identity)
    curve(out, 'cube',   1, colors[1], styles[1])
    data (out, 'n', 20000, dimlist, 10, 0, 0, identity)
    curve(out, 'normal', 0, colors[0], styles[0])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+2) +');\n')
    out.write('  init_hubs(21.0, '
             +'btex 10 nearest neighbors, 10000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex measure $h_1(1\\mbox{\\scriptsize \\%})$ etex);\n')
    data (out, 's', 10000, spdlist, 10, 0, 4, minus1)
    curve(out, 'sphere', 3, colors[3], styles[3])
    data (out, 'b', 10000, dimlist, 10, 0, 4, identity)
    curve(out, 'ball',   2, colors[2], styles[2])
    data (out, 'c', 10000, dimlist, 10, 0, 4, identity)
    curve(out, 'cube',   1, colors[1], styles[1])
    data (out, 'n', 10000, dimlist, 10, 0, 4, identity)
    curve(out, 'normal', 0, colors[0], styles[0])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+3) +');\n')
    out.write('  init_hubs(21.0, '
             +'btex 10 nearest neighbors, 20000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex measure $h_1(1\\mbox{\\scriptsize \\%})$ etex);\n')
    data (out, 's', 20000, spdlist, 10, 0, 4, minus1)
    curve(out, 'sphere', 3, colors[3], styles[3])
    data (out, 'b', 20000, dimlist, 10, 0, 4, identity)
    curve(out, 'ball',   2, colors[2], styles[2])
    data (out, 'c', 20000, dimlist, 10, 0, 4, identity)
    curve(out, 'cube',   1, colors[1], styles[1])
    data (out, 'n', 20000, dimlist, 10, 0, 4, identity)
    curve(out, 'normal', 0, colors[0], styles[0])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+4) +');\n')
    out.write('  init_hubs(8.5, '
             +'btex 10 nearest neighbors, 10000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex measure $h_2(3)$ in \% etex);\n')
    data (out, 's', 10000, spdlist, 10, 0, 6, minus1)
    curve(out, 'sphere', 3, colors[3], styles[3])
    data (out, 'b', 10000, dimlist, 10, 0, 6, identity)
    curve(out, 'ball',   2, colors[2], styles[2])
    data (out, 'c', 10000, dimlist, 10, 0, 6, identity)
    curve(out, 'cube',   1, colors[1], styles[1])
    data (out, 'n', 10000, dimlist, 10, 0, 6, identity)
    curve(out, 'normal', 0, colors[0], styles[0])
    legend(out, (-0.5,-9.0), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+5) +');\n')
    out.write('  init_hubs(8.5, '
             +'btex 10 nearest neighbors, 20000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex measure $h_2(3)$ in \% etex);\n')
    data (out, 's', 20000, spdlist, 10, 0, 6, minus1)
    curve(out, 'sphere', 3, colors[3], styles[3])
    data (out, 'b', 20000, dimlist, 10, 0, 6, identity)
    curve(out, 'ball',   2, colors[2], styles[2])
    data (out, 'c', 20000, dimlist, 10, 0, 6, identity)
    curve(out, 'cube',   1, colors[1], styles[1])
    data (out, 'n', 20000, dimlist, 10, 0, 6, identity)
    curve(out, 'normal', 0, colors[0], styles[0])
    legend(out, (-0.5,-9.0), '')
    out.write('endfig;\n\n')

#-----------------------------------------------------------------------

def jolt_diags(out, fid):
    out.write('vardef init_jolt (expr ht, txt) =\n')
    out.write('  pair data[];\n')
    out.write('  linechart.set_defaults("ptmr8r");\n')
    out.write('  linechart.legend.linelen  := 3mm;\n')
    out.write('  linechart.legend.itemskip := '
              + '0.9linechart.legend.itemskip;\n')
    out.write('  linechart.set_units(44mm/104, 26mm/ht);\n')
    out.write('  linechart.set_ranges((0,104), (0,ht));\n')
    out.write('  linechart.draw_xticks(0, 20, 40, 60, 80, 100);\n')
    out.write('  linechart.draw_xlabel(0, '
                   +'btex number of dimensions $n$ etex);\n')
    out.write('  if     ht > 12:\n');
    out.write('    linechart.y.alabel.offset := (0,5.5mm);\n')
    out.write('    linechart.draw_yticks(0, 5, 10, 15, 20, 25);\n')
    out.write('  elseif ht > 6:\n')
    out.write('    linechart.y.alabel.offset := (0,3.2mm);\n')
    out.write('    linechart.draw_yticks(0, 2, 4, 6, 8, 10, 12);\n')
    out.write('  else:\n')
    out.write('    linechart.y.alabel.offset := (0,3.2mm);\n')
    out.write('    linechart.draw_yticks(0, 1, 2, 3, 4, 5, 6);\n')
    out.write('  fi\n')
    out.write('  linechart.draw_frame;\n')
    out.write('  linechart.draw_title(2, txt);\n')
    out.write('enddef;\n\n')

    dimlist = [ 2, 3, 5, 10, 15, 20, 30, 40, 50, 60, 80, 100 ]

    out.write('beginfig(' +str(fid) +');\n')
    out.write('  init_jolt(5.95, '
             +'btex 10 nearest neighbors, 10000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex skewness $\gamma$ etex);\n')
    data (out, 'c', 10000, dimlist, 10, 0, 0, identity)
    curve(out, 'cube',   0, colors[4], styles[4])
    data (out, 'b', 10000, dimlist, 10, 0, 0, identity)
    curve(out, 'ball',   1, colors[5], styles[5])
    data (out, 'j', 10000, dimlist, 10, 0, 0, identity)
    curve(out, 'jolted', 2, colors[6], styles[6])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+1) +');\n')
    out.write('  init_jolt(5.95, '
             +'btex 10 nearest neighbors, 20000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex skewness $\gamma$ etex);\n')
    data (out, 'c', 20000, dimlist, 10, 0, 0, identity)
    curve(out, 'cube',   0, colors[4], styles[4])
    data (out, 'b', 20000, dimlist, 10, 0, 0, identity)
    curve(out, 'ball',   1, colors[5], styles[5])
    data (out, 'j', 20000, dimlist, 10, 0, 0, identity)
    curve(out, 'jolted', 2, colors[6], styles[6])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+2) +');\n')
    out.write('  init_jolt(11.0, '
             +'btex 10 nearest neighbors, 10000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex measure $h_1(1\\mbox{\\scriptsize \\%})$ etex);\n')
    data (out, 'c', 10000, dimlist, 10, 0, 4, identity)
    curve(out, 'cube',   0, colors[4], styles[4])
    data (out, 'b', 10000, dimlist, 10, 0, 4, identity)
    curve(out, 'ball',   1, colors[5], styles[5])
    data (out, 'j', 10000, dimlist, 10, 0, 4, identity)
    curve(out, 'jolted', 2, colors[6], styles[6])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+3) +');\n')
    out.write('  init_jolt(11.0, '
             +'btex 10 nearest neighbors, 20000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex measure $h_1(1\\mbox{\\scriptsize \\%})$ etex);\n')
    data (out, 'c', 20000, dimlist, 10, 0, 4, identity)
    curve(out, 'cube',   0, colors[4], styles[4])
    data (out, 'b', 20000, dimlist, 10, 0, 4, identity)
    curve(out, 'ball',   1, colors[5], styles[5])
    data (out, 'j', 20000, dimlist, 10, 0, 4, identity)
    curve(out, 'jolted', 2, colors[6], styles[6])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+4) +');\n')
    out.write('  init_jolt(7.1, '
             +'btex 10 nearest neighbors, 10000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex measure $h_2(3)$ in \% etex);\n')
    data (out, 'c', 10000, dimlist, 10, 0, 6, identity)
    curve(out, 'cube',   0, colors[4], styles[4])
    data (out, 'b', 10000, dimlist, 10, 0, 6, identity)
    curve(out, 'ball',   1, colors[5], styles[5])
    data (out, 'j', 10000, dimlist, 10, 0, 6, identity)
    curve(out, 'jolted', 2, colors[6], styles[6])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+5) +');\n')
    out.write('  init_jolt(7.1, '
             +'btex 10 nearest neighbors, 20000 points etex);\n')
    out.write('  linechart.draw_ylabel(0, '
             +'btex measure $h_2(3)$ in \% etex);\n')
    data (out, 'c', 20000, dimlist, 10, 0, 6, identity)
    curve(out, 'cube',   0, colors[4], styles[4])
    data (out, 'b', 20000, dimlist, 10, 0, 6, identity)
    curve(out, 'ball',   1, colors[5], styles[5])
    data (out, 'j', 20000, dimlist, 10, 0, 6, identity)
    curve(out, 'jolted', 2, colors[6], styles[6])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

#-----------------------------------------------------------------------

def cube_diags (out, fid):
    out.write('vardef init_cubes (expr ht, csize, txt) =\n')
    out.write('  pair data[];\n')
    out.write('  linechart.set_defaults("ptmr8r");\n')
    out.write('  linechart.legend.linelen  := 3mm;\n')
    out.write('  linechart.legend.itemskip := '
              + '0.9linechart.legend.itemskip;\n')
    out.write('  linechart.y.alabel.offset := (0,3.0mm);\n')
    out.write('  if csize:\n')
    out.write('    linechart.set_units(44mm/52, 26mm/ht);\n')
    out.write('    linechart.set_ranges((0,52), (0,ht));\n')
    out.write('    linechart.draw_xticks(0, 10, 20, 30, 40, 50);\n')
    out.write('    linechart.draw_xlabel(0, '
                     +'btex percentage of gap per grid cell etex);\n')
    out.write('  else:\n')
    out.write('    linechart.set_units(44mm/9.6, 26mm/ht);\n')
    out.write('    linechart.set_ranges((0.7,10.3), (0,ht));\n')
    out.write('    linechart.draw_xticks(1, 2, 4, 6, 8, 10);\n')
    out.write('    linechart.draw_xlabel(1, '
                     +'btex number $\eta$ of cubes per dimension etex);\n')
    out.write('  fi\n')
    out.write('  linechart.draw_yticks(0, 1, 2, 3, 4, 5);\n')
    out.write('  linechart.draw_ylabel(0, '
                   +'btex measure $h_2(2)$ in \% etex);\n')
    out.write('  linechart.draw_frame;\n')
    out.write('  linechart.draw_title(1, txt);\n')
    out.write('enddef;\n\n')

    cubelist = [0.0, 2.5, 4.5, 6.5, 8.5, 10.5]
    sizelist = [0.0, 8.9, 8.8, 8.7, 8.6, 8.5]

    out.write('beginfig(' +str(fid) +');\n')
    out.write('  init_cubes(5.0, false, '
             +'btex 5 nearest neighbors etex);\n')
    data (out, 'c', 10000, 3, 5, cubelist, 7, cubes)
    curve(out, '10000 points', 0, colors[7], styles[7])
    data (out, 'c', 20000, 3, 5, cubelist, 7, cubes)
    curve(out, '20000 points', 1, colors[8], styles[8])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+1) +');\n')
    out.write('  init_cubes(3.3, false, '
             +'btex 8 nearest neighbors etex);\n')
    data (out, 'c', 10000, 3, 8, cubelist, 7, cubes)
    curve(out, '10000 points', 0, colors[7], styles[7])
    data (out, 'c', 20000, 3, 8, cubelist, 7, cubes)
    curve(out, '20000 points', 1, colors[8], styles[8])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+2) +');\n')
    out.write('  init_cubes(5.0, true, '
             +'btex 5 nearest neighbors etex);\n')
    data (out, 'c', 10000, 3, 5, sizelist, 7, csize)
    curve(out, '10000 points', 0, colors[7], styles[7])
    data (out, 'c', 20000, 3, 5, sizelist, 7, csize)
    curve(out, '20000 points', 1, colors[8], styles[8])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

    out.write('beginfig(' +str(fid+3) +');\n')
    out.write('  init_cubes(3.3, true, '
             +'btex 8 nearest neighbors etex);\n')
    data (out, 'c', 10000, 3, 8, sizelist, 7, csize)
    curve(out, '10000 points', 0, colors[7], styles[7])
    data (out, 'c', 20000, 3, 8, sizelist, 7, csize)
    curve(out, '20000 points', 1, colors[8], styles[8])
    legend(out, (0.5,-0.5), '')
    out.write('endfig;\n\n')

#-----------------------------------------------------------------------

if __name__ == '__main__':
    if len(argv) > 1:           # if to use color
        colors = [ 'red', '0.7green', 'blue', 'black',
                   '0.7green', 'blue', 'red',
                   'blue', 'red' ]
        styles = [ '', '', '', '', '', '', '', '', '' ]
    else:                       # if to use only gray scales
        colors = [ '0.95white', '0.75white', '0.5white', '0.2white',
                   '0.75white', '0.5white', '0.2white',
                   '0.2white', '0.6white' ]
        styles = [ '', '', '', '', '', '', '', '', '' ]
    with open('diags.mp', 'w') as out:
        header(out, 3)
        hubs_diags(out,  0)
        jolt_diags(out,  6)
        cube_diags(out, 12)
        footer(out)
